IMPLEMENTATION MODULE ShdMem;

IMPORT WinDef, WinNT, WinBase;
IMPORT IO, Lib;
IMPORT Types, SO, Cfg, Log, Config;
FROM SYSTEM IMPORT BYTE, ADDRESS, ADR;

CONST
  mid = Cfg.midShdMem;
  nam = 'Local\ShdMemTst';  (* 'Global\' seems to require admin rights *)
  namMtx = 'Local\ShMemTst_Mtx';

TYPE
  THandle = WinDef.HANDLE;

CONST
  hndInv = WinBase.INVALID_HANDLE_VALUE;  (* 0FFFFFFFFH *)

VAR
  adr: ADDRESS;
  mtx: THandle;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, FALSE );
END Message;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE WrHex( VAR b: SO.TBuffer; x-: ARRAY OF BYTE; l: CARDINAL );
CONST
  lMax = 8;
VAR
  i: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > lMax THEN l := lMax END;
  IF l = 0 THEN RETURN END;
  c := b.c.p;
  b.c.p := '0';
  DEC( l );
  FOR i := 0 TO l DO
    SO.WrShtHex( b, Types.TShtCard( x[i] ), 2 );
    IF i < l THEN SO.WrChar( b, ' ' ) END;
  END;
  b.c.p := c;
END WrHex;


PROCEDURE NewMtx( n-: ARRAY OF CHAR ): THandle;
VAR
  h: THandle;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateMutex( NIL, FALSE, n );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewMtx: CreateMutex failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  RETURN h;
END NewMtx;


(* the shared memory is not automatically interlocked by win (tested) *)
PROCEDURE Enter();
VAR
  i: LONGCARD;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ Config.data.mtx THEN RETURN END;
  i := WinBase.WaitForSingleObject( mtx, WinBase.INFINITE );
  IF i = MAX( LONGCARD ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  IF i > 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject returned ' );
    SO.WrLngCard( b, i, 1 );
    Fatal( s );
  END;
END Enter;

PROCEDURE Leave();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ Config.data.mtx THEN RETURN END;
  IF ~ WinBase.ReleaseMutex( mtx ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Leave: ReleaseMutex failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END Leave;


PROCEDURE Write( x-: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Write: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > len THEN l := len END;
  Enter();
  Lib.Move( ADR( x ), adr, l );
  Leave();
  Show();
  RETURN l;
END Write;

PROCEDURE Read( VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Read: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > len THEN l := len END;
  Enter();
  Lib.Move( adr, ADR( x ), l );
  Leave();
  Show();
  RETURN l;
END Read;


PROCEDURE Create();
VAR
  h: THandle;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateFileMapping(
    hndInv,                (* file handle *)
    NIL,                   (* attributes *)
    WinNT.PAGE_READWRITE,  (* protection *)
    0,                     (* size high *)
    len,                   (* size low *)
    nam                    (* name *)
  );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Create: CreateFileMapping failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;

  (* WinError.ERROR_ALREADY_EXISTS = 183 *)
  SO.Init( b, s );
  SO.WrStr( b, 'Create: CreateFileMapping succeeded, error code ' );
  SO.WrLngCard( b, WinBase.GetLastError(), 1 );
  Message( s );

  adr := WinBase.MapViewOfFile(
    h,
    WinBase.FILE_MAP_ALL_ACCESS,
    0,  (* offset high *)
    0,  (* offset low *)
    len (* size (still unclear, since 1 works) *)
  );
  IF adr = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Create: MapViewOfFile failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END Create;

PROCEDURE InitModule();
BEGIN
  Log.Register( mid, 'ShdMem' );
  Create();
  mtx := NewMtx( namMtx );
END InitModule;

BEGIN
  InitModule();
END ShdMem.
